from django_pds.conf import settings
from django_pds.core.controllers import GenericUpdateCommandController

SECURITY_ATTRIBUTES = settings.SECURITY_ATTRIBUTES
READ_ONLY_FIELDS = settings.READ_ONLY_FIELDS


def data_update(document_name, data, user_id=None, ignore_security=False):
    try:

        update_ctrl = GenericUpdateCommandController()

        if ignore_security:
            return update_ctrl.update_one(document_name, data)

        data_fields = set(data.keys())
        security_attributes = set(SECURITY_ATTRIBUTES)
        common_fields = update_ctrl.common_fields(data_fields, security_attributes)

        if len(common_fields) > 0:
            fields = ','.join(common_fields)
            return True, f'the following security attributes {fields} found in the json data'

        read_only_fields = set(READ_ONLY_FIELDS)
        rof = update_ctrl.common_fields(data_fields, read_only_fields)
        if len(rof) > 0:
            fields = ','.join(rof)
            return True, f'the following read only attributes {fields} found in the json data'

        can_update = update_ctrl.can_update(document_name, data.get('ItemId', None), user_id)

        if can_update:
            return update_ctrl.update_one(document_name, data, user_id)
        else:
            return True, "access denied, you don't have sufficient permission to update"
    except BaseException as e:
        return True, str(e)



from .delete import data_delete
from .read import data_read, basic_data_read
from .update import data_update
from .upsert import data_upsert
from .write import data_insert

__all__ = (
    'data_read',
    'basic_data_read',
    'data_insert',
    'data_update',
    'data_upsert',
    'data_delete'
)



from mongoengine import StringField

from django_pds.core.base import BaseDocument


class UserRoleMap(BaseDocument):

    RoleName = StringField(required=True)
    RoleId = StringField(required=False, max_length=36)

    meta = {'collection': 'UserRoleMaps'}



from django.http import JsonResponse


def server_error(request, *args, message=None, **kwargs):
    message = message or "Server has encountered an error 500"
    """
    Generic 500 error handler.
    """
    res = {
        "error_status": 500,
        "error": "server_error",
        "error_description": message
    }
    return JsonResponse(res, status=500)


def access_denied(request, message=None):
    if not message:
        message = "You don't have sufficient permission to access this specified HTTP URL"
    res = {
        "error_status": 401,
        "error": "access_permission_error",
        "error_description": message
    }
    return JsonResponse(res, status=401)


def url_not_found(request):
    res = {
        "error": "invalid_api_url",
        "error_status": 404,
        "error_message": "The specified HTTP URL is not valid."
    }
    return JsonResponse(res, status=404)


def method_not_allowed():
    res = {
        "error_status": 405,
        "error": "invalid_method",
        "error_description": "The specified HTTP method is not valid."
    }
    return JsonResponse(res, status=405)



from mongoengine import Q

from django_pds.conf import settings
from django_pds.core.controllers import UserReadableDataController, GenericReadController, UserRoleMapsController
from django_pds.core.rest.response import error_response, success_response_with_total_records
from django_pds.core.utils import get_fields, get_document, is_abstract_document
from django_pds.core.utils import print_traceback
from django_pds.serializers import GenericSerializerAlpha
from ..parser.query import QueryParser
from ..parser.terms import FILTER, WHERE, SELECT, PAGE_SIZE, PAGE_NUM, ORDER_BY, RAW_WHERE

NOT_SELECTABLE_ENTITIES_BY_PDS = settings.SELECT_NOT_ALLOWED_ENTITIES
SECURITY_ATTRIBUTES = settings.SECURITY_ATTRIBUTES


def basic_data_read(document_name, fields='__all__',
                    page_size=10, page_num=1, order_by=None,
                    include_security_fields=False,
                    error_track=False):
    try:

        document = get_document(document_name)

        if not document or not document_name:
            return True, error_response(f'document by name `{document_name}` doesn\'t exists')

        if fields != '__all__' and not isinstance(fields, (list, tuple)):
            return True, error_response('fields must be a list or tuple')

        sql_ctrl = GenericReadController()
        data, cnt = sql_ctrl.read(document_name, Q(), page_size, page_num, order_by)
        if cnt == 0:
            return False, success_response_with_total_records([], cnt)
        gsa = GenericSerializerAlpha(document_name=document_name)
        if not fields == '__all__':
            for field in fields:
                gsa.select(field)
        else:
            fields = get_fields(document_name)
            if not include_security_fields:
                fields = tuple(set(fields) - set(SECURITY_ATTRIBUTES))
            gsa.fields(fields)

        json = gsa.serialize(data)
        res = success_response_with_total_records(json.data, cnt)
        return False, res
    except BaseException as e:
        if error_track:
            print_traceback()
        return True, error_response(str(e))


def data_read(
        document_name, sql_text, user_id=None,
        roles=None, checking_roles=True,
        readable=True, security_attributes=True,
        selectable=True, read_all=False, exclude_default=False,
        page_number=1, _size=10, error_track=False):
    """
    :param page_number:
    :param _size:
    :param checking_roles:
    :param document_name:
    :param sql_text:
    :param user_id:
    :param roles:
    :param readable:
    :param security_attributes:
    :param selectable:
    :param read_all:
    :return:
    """

    document = get_document(document_name)

    # checking either model exists
    # or entity exists in not selectable entities

    if not document:
        return False, error_response('document model not found')

    if is_abstract_document(document_name):
        return False, error_response('document model not found')

    if selectable and document_name in NOT_SELECTABLE_ENTITIES_BY_PDS:
        return False, error_response('document model is not selectable')

    try:
        parser = QueryParser(sql_text)
        dictionary = parser.parse()

        # filtering fields in where clause
        _filters = []
        if dictionary.get(FILTER, None):
            _filters = dictionary[FILTER]

        filter_fields = set(_filters)
        document_fields = set(get_fields(document_name))

        if len(filter_fields - document_fields) > 0:
            return True, error_response('Where clause contains unknown attribute to this Document')

        if security_attributes:
            security_attr = set(SECURITY_ATTRIBUTES)
            contains_security_attributes = filter_fields.intersection(security_attr)
            if len(contains_security_attributes) > 0:
                return True, error_response('Security attributes found in where clause')

        # checking user readable data from database for this particular request

        fields = ['ItemId']
        if dictionary.get(SELECT, None):
            fields = dictionary[SELECT]
        if read_all:
            fields = document_fields

        urm_ctrl = UserRoleMapsController()

        if readable:
            urds_ctrl = UserReadableDataController()
            __roles = None
            if user_id and not roles:
                __roles = urm_ctrl.get_user_roles(user_id)
            err, _fields = urds_ctrl.get_user_readable_data_fields(document_name, __roles, exclude_default)
            if err:
                msg = f'Entity \'{document_name}\' is missing from user readable data\'s'
                return True, error_response(msg)

            diff = set(fields) - _fields  # _fields are already a set
            if len(diff) > 0:
                return True, error_response("Select clause contains non readable attributes")

        sql_ctrl = GenericReadController()
        __raw__where = dictionary.get(RAW_WHERE, {})

        page_num = dictionary.get(PAGE_NUM, page_number)
        page_size = dictionary.get(PAGE_SIZE, _size)

        q = Q()
        if dictionary.get(WHERE, None):
            q = dictionary[WHERE]

        # checking for row level permission starts

        q2 = Q()

        if user_id:
            q2 = Q(IdsAllowedToRead=user_id)

        if checking_roles:
            if not roles and user_id:
                roles = urm_ctrl.get_user_roles(user_id)
            if roles and not isinstance(roles, (list, tuple)):
                return True, error_response('roles must be a list or a tuple.')
            for role in roles:
                q2 = q2.__or__(Q(RolesAllowedToRead=role))

        if user_id or (checking_roles and roles):
            q = q.__and__(q2)

        # checking for row level permission ends

        order_by = []
        if dictionary.get(ORDER_BY, None):
            order_by = dictionary[ORDER_BY]

        data, cnt = sql_ctrl.read(document_name, q, page_size, page_num, order_by)
        if cnt == 0:
            return False, success_response_with_total_records([], cnt)
        gsa = GenericSerializerAlpha(document_name=document_name)
        for field in fields:
            gsa.select(field)
        json = gsa.serialize(data)
        res = success_response_with_total_records(json.data, cnt)
        return False, res
    except BaseException as e:
        if error_track:
            print_traceback()
        return True, error_response(str(e))



from django_pds.conf import settings
from django_pds.core.controllers import DefaultPermissionSettingsController
from django_pds.core.controllers import GenericInsertCommandController

security_attributes = settings.SECURITY_ATTRIBUTES
read_only_fields = settings.READ_ONLY_FIELDS


def data_insert(document_name, data, user_id=None, ignore_security=False, force_insert=False):

    insert_ctrl = GenericInsertCommandController()

    if force_insert:
        already_exists = insert_ctrl.already_exists(document_name, data.get('ItemId', None))
        if already_exists and not force_insert:
            return True, "document ItemId already exists, you can't create new collection " \
                         "with the same ItemId, if you want to update, " \
                         "use pds update or upsert method"

    if ignore_security:
        return insert_ctrl.insert_one(document_name, data)

    entity_permission = DefaultPermissionSettingsController()
    can_insert = entity_permission.can_insert(document_name, user_id)

    if can_insert:

        data_fields = set(data.keys())
        rof = set(read_only_fields)
        common_fields = data_fields.intersection(rof)
        if len(common_fields) > 0:
            fields = ','.join(common_fields)
            return True, f'the following read only attributes {fields} found in the json data'
        sec_attr = set(security_attributes)
        common_fields = data_fields.intersection(sec_attr)
        if len(common_fields) > 0:
            fields = ','.join(common_fields)
            return True, f'following security attributes {fields} found in the json data'
        # insert data after all the verification and security checking's
        error, permissions = entity_permission.get_document_name_permissions(document_name)
        if error:
            return True, error
        return insert_ctrl.insert_one(document_name, data, user_id, permissions)
    else:
        return True, "access denied, you don't have sufficient permission to insert data"



KWARGS_BUILDER_SECRET = 'FXV/#X=>fMT,pc-wm3BYaxqoZ7VOA+'


class KwargsBuilder:

    def purify(self, **kwargs):
        _dict = {}
        for key, value in kwargs.items():
            if value is not KWARGS_BUILDER_SECRET:
                _dict[key] = value
        return _dict

    def build(self, _dict, data, data_key, dict_key=None):
        if dict_key:
            _dict[dict_key] = data.get(data_key, KWARGS_BUILDER_SECRET)
        else:
            _dict[data_key] = data.get(data_key, KWARGS_BUILDER_SECRET)



from .base import BaseController, RequiredController
from .delete import GenericDeleteCommandController
from .entity import DefaultPermissionSettingsController
from .insert import GenericInsertCommandController
from .kwargs import KwargsBuilder
from .read import GenericReadController
from .role import RoleController
from .update import GenericUpdateCommandController
from .userreadabledata import UserReadableDataController
from .userrolemaps import UserRoleMapsController

__all__ = [
    'BaseController',
    'RequiredController',
    'GenericDeleteCommandController',
    'DefaultPermissionSettingsController',
    'GenericInsertCommandController',
    'KwargsBuilder',
    'GenericReadController',
    'RoleController',
    'GenericUpdateCommandController',
    'UserReadableDataController',
    'UserRoleMapsController'
]



from mongoengine import StringField, ListField

from django_pds.core.base import SimpleBaseDocument


class EntityDefaultPermissionSetting(SimpleBaseDocument):
    EntityName = StringField(required=True)
    IdsAllowedToRead = ListField(StringField(max_length=36), default=[])
    IdsAllowedToWrite = ListField(StringField(max_length=36), default=[])
    IdsAllowedToUpdate = ListField(StringField(max_length=36), default=[])
    IdsAllowedToDelete = ListField(StringField(max_length=36), default=[])
    RolesAllowedToRead = ListField(StringField(max_length=36), default=[])
    RolesAllowedToWrite = ListField(StringField(max_length=36), default=[])
    RolesAllowedToUpdate = ListField(StringField(max_length=36), default=[])
    RolesAllowedToDelete = ListField(StringField(max_length=36), default=[])

    meta = {'collection': 'EntityDefaultPermissionSettings'}



from mongoengine import StringField

from django_pds.core.base import SimpleBaseDocument


class Entity(SimpleBaseDocument):
    EntityName = StringField(max_length=120, required=True, null=False)
    PrimaryEntityName = StringField(max_length=120, required=True, null=False)

    meta = {'collection': 'Entities'}



from mongoengine import StringField, ListField, BooleanField

from django_pds.core.base import SimpleBaseDocument


class Role(SimpleBaseDocument):
    RoleName = StringField(required=True)
    ParentRoles = ListField(StringField())
    IsDynamic = BooleanField(default=False)

    meta = {'collection': 'Roles'}



from django_pds.core.controllers import DefaultPermissionSettingsController, GenericDeleteCommandController


def data_delete(document_name, document_id, user_id=None, ignore_permissions=False):
    try:

        delete_ctrl = GenericDeleteCommandController()

        if ignore_permissions:
            return delete_ctrl.delete(document_name, document_id)

        entity_permission = DefaultPermissionSettingsController()
        role_can_delete = False
        if user_id:
            role_can_delete = entity_permission.can_delete(document_name, user_id)
        id_can_delete = False
        if not role_can_delete and not user_id:
            id_can_delete = delete_ctrl.has_permission(document_name, user_id, document_id)
        if role_can_delete or id_can_delete:
            return delete_ctrl.delete(document_name, document_id)
        else:
            return True, "you don't have sufficient permission to delete"
    except BaseException as e:
        return True, str(e)



import jwt
from django.core.cache import cache
from django.middleware.http import MiddlewareMixin

from django_pds.conf import settings
from django_pds.core.rest.exceptions import access_denied, server_error
from django_pds.core.utils import authorization_token, path
from .settings import conf

TOKEN_MISSING = 'Authorization token not found in the request headers'
INVALID_TOKEN = 'Invalid authentication token'
ACCESS_DENIED = 'Access denied'
TOKEN_AUTH_KEY_NOT_SET = 'Token secret key not set'

audience = settings.JWT_TOKEN_AUDIENCE
algorithm = settings.JWT_TOKEN_ALGORITHM
auth_key = getattr(settings, 'JWT_TOKEN_SECRET_KEY', None)


class AuthenticationMiddleware(MiddlewareMixin):

    def process_request(self, request):

        url = path(request)

        if not auth_key:
            return server_error(request, message=TOKEN_AUTH_KEY_NOT_SET)

        bearer_token = authorization_token(request)

        flag = True

        for exempt_url in conf.SESSION_MIDDLEWARE_EXEMPT_URLS:
            if url.endswith(exempt_url):
                flag = False

        if not flag:
            return None

        if bearer_token is None:
            return access_denied(request, TOKEN_MISSING)

        try:

            token_hash = str(hash(bearer_token))
            payload = cache.get(token_hash)

            if payload is None:
                token = bearer_token.split()
                payload = jwt.decode(token[1], auth_key, audience=audience, unicode='utf-8', algorithms=algorithm)
                if payload is None:
                    return access_denied(request, message=INVALID_TOKEN)

            auth = payload.get('logged_in', False)

            for exempt_url in conf.AUTH_NOT_REQUIRED_URLS:
                if url.endswith(exempt_url):
                    auth = True

            if auth:
                return None
            else:
                return access_denied(request, ACCESS_DENIED)
        except BaseException as e:
            return access_denied(request, str(e))



import json

from django.utils.timezone import now

from django_pds.core.controllers.base import BaseController
from django_pds.core.utils import get_document
from .userrolemaps import UserRoleMapsController

ADMIN = "admin"


class GenericUpdateCommandController(BaseController):

    def json_load(self, json_string):
        try:
            return False, json.loads(json_string)
        except BaseException as e:
            return True, str(e)

    def already_exists(self, document_name, item_id):
        try:
            data = get_document(document_name).objects(ItemId=item_id)
            return data.count() > 0
        except BaseException as e:
            return False

    def can_update(self, document_name, item_id, user_id):

        data = get_document(document_name).objects(ItemId=item_id)
        if data.count() > 0:
            data = data[0]
            user_role_ctrl = UserRoleMapsController()
            user_roles = set(user_role_ctrl.get_user_roles(user_id))
            permitted_roles = set(getattr(data, 'RolesAllowedToUpdate', []))
            permitted_ids = set(getattr(data, 'IdsAllowedToUpdate', []))
            permitted_roles.add(ADMIN)
            common_roles = user_roles.intersection(permitted_roles)
            return len(common_roles) > 0 or user_id in permitted_ids
        return False

    def update_one(self, document_name, data, user_id=None):
        try:

            base_instance = self.is_base_instance(document_name)
            simple_base_instance = self.is_simple_base_doc_instance(document_name)

            if not base_instance and not simple_base_instance:
                return True, 'Document type must be `BaseDocument` ' \
                             'or `SimpleBaseDocument` ' \
                             'from django_pds.core.base Module'

            Model = get_document(document_name)
            item_id = data.get('ItemId', None)
            _data = Model.objects(ItemId=item_id)[0]
            for field in data:
                setattr(_data, field, data.get(field))

            if base_instance:
                _data.LastUpdateDate = now()
                if user_id:
                    _data.LastUpdateBy = user_id
            _data.save()
            return False, _data.ItemId

        except BaseException as e:
            return True, e

    def update_many(self, document_name, data_array, user_id=None):
        results = []
        for data in data_array:
            err, item_id = self.update_one(document_name, data, user_id)
            if err:
                results.append(None)
            else:
                results.append(item_id)
        return results



import json
import re

from mongoengine import Q

from django_pds.core.helper.xstr import xstr
from .terms import WHERE, SELECT, EQUAL, LT, GT, AND, OR, EMPTY, COMMA, FILTER, PARENTHESIS_LEFT, PARENTHESIS_RIGHT, \
    ORDER_BY, RAW_WHERE

REGEX = r'\(|\)|\w+=[\w+|\-*|\d*|\:*|\.*|\ *|\,*|\'*|\%*]*|\&+|\@+'

REGEX_STRING_QUERY = ['exact', 'iexact', 'contains', 'icontains', 'startswith', 'istartswith',
                      'endswith', 'iendswith', 'match']

REPLACE_DICT = {
    '%3D': '=',
    '%2B': '+',
    '%2F': '/',
    '%20': ' ',
    '%25': '%'
}


class QueryParser:
    __raw = {}

    def __init__(self, text):
        self.text = text

    def __count_lt_gt(self):
        cnt_lt = self.text.count(LT)
        cnt_gt = self.text.count(GT)
        return cnt_lt == cnt_gt

    def __count_parenthesis(self):
        cnt_lp = self.text.count(PARENTHESIS_LEFT)
        cnt_rp = self.text.count(PARENTHESIS_RIGHT)
        return cnt_lp == cnt_rp

    def validate(self):
        if not self.__count_lt_gt():
            raise Exception('Your SQL missing some < / >')
        if not self.__count_parenthesis():
            raise Exception('Your SQL missing some ( / )')

    def find(self, string, ch):
        pos = []
        ln = len(string)
        for i in range(ln):
            if string[i] in ch:
                pos.append(i)
        return pos

    def replacer(self, clause):
        data = ''
        for i in range(len(clause)):
            if clause[i] == '|':
                data = data + '@'
            elif clause[i] == '@':
                data = data + '|'
            else:
                data = data + clause[i]
        return data

    def reverse_replacer(self, clause):
        return self.replacer(clause)

    def deconstruct_select(self, clause):
        fields = clause.split(COMMA)
        for i in range(len(fields)):
            fields[i] = fields[i].strip()
        return fields

    def make_tree(self, data):
        items = re.findall(REGEX, data)

        def req(index):
            result = []
            item = items[index]
            while item != ")":
                if item == "(":
                    subtree, index = req(index + 1)
                    result.append(subtree)
                else:
                    result.append(item)
                index += 1
                item = items[index]
            return result, index

        return req(1)[0]

    def equal_splitter(self, clause):
        items = clause.split("=")
        return items[0], xstr(items[1]).get()

    def print_dumps(self, dictionary):
        y = json.dumps(dictionary, indent=4, sort_keys=True)
        print(y)

    def making_q(self, tree, _filter, q, start, position=None):
        if position:
            subtree = tree[start:position]
        else:
            subtree = tree[start:]
        if len(subtree) > 1:
            return None
        if isinstance(subtree[0], list):
            sub_q = self.construct_q(subtree[0], _filter)
            if sub_q and start == 0:
                q = sub_q
            elif sub_q and tree[start - 1] == OR:
                q = q.__or__(sub_q)
            elif sub_q and tree[start - 1] == AND:
                q = q.__and__(sub_q)
        else:
            key, value = self.equal_splitter(subtree[0])
            keys = key.strip().rsplit("__", 1)
            _filter.append(keys[0].strip())
            for replace_with in REPLACE_DICT:
                value = value.replace(replace_with, REPLACE_DICT.get(replace_with))
            if len(keys) > 1:
                if keys[1] in REGEX_STRING_QUERY:
                    value = str(value)
            value = self.replacer(value)
            self.__raw[keys[0].strip()] = value
            _pair = {key.strip(): value}
            if start == 0:
                q = Q(**_pair)
            else:
                if tree[start - 1] == AND:
                    q = q.__and__(Q(**_pair))
                elif tree[start - 1] == OR:
                    q = q.__or__(Q(**_pair))

        return q

    def construct_q(self, tree, _filter):
        operators = ['&', '@']
        positions = list(self.find(tree, operators))
        start = 0
        q = Q()
        for position in positions:
            q = self.making_q(tree, _filter, q, start, position)
            start = position + 1
        q = self.making_q(tree, _filter, q, start)
        return q

    def deconstruct_deep_where(self, clause):
        if not clause.endswith(")") or not clause.startswith("("):
            clause = "(" + clause + ")"
        tree = self.make_tree(clause)
        _filter = []
        q = self.construct_q(tree, _filter)
        return _filter, q

    def parse(self):
        self.validate()
        splits = self.text.split(GT)
        dictionary = {}
        for keyword in splits:
            keywords = keyword.split(LT)

            if len(keywords) == 2:
                key = keywords[0].strip()
                key = key.replace(EQUAL, EMPTY)
                at = xstr(keywords[1].strip())
                dictionary[key.lower()] = at.get()

        if dictionary.get(ORDER_BY, None):
            clause = dictionary[ORDER_BY]
            data = []
            items = clause.split(',')
            for item in items:
                data.append(item.strip())
            dictionary[ORDER_BY] = data

        if dictionary.get(SELECT, None):
            dictionary[SELECT] = self.deconstruct_select(dictionary[SELECT])
        if dictionary.get(WHERE, None):
            clause = self.replacer(dictionary[WHERE])
            # pre-caution
            # or operator replaced with @ because | doesn't work
            # with regex that well
            _filters, _where = self.deconstruct_deep_where(clause)
            dictionary[WHERE] = _where
            dictionary[FILTER] = _filters
            dictionary[RAW_WHERE] = self.__raw
        return dictionary



DEFAULT_USER_ROLE = 'user'

SYSTEM_SUPPORTED_ROLES = [
    "admin",
    "user",
    "anonymous"
]

SECURITY_ATTRIBUTES = [
    'IdsAllowedToRead', 'IdsAllowedToWrite', 'IdsAllowedToUpdate', 'IdsAllowedToDelete',
    'RolesAllowedToRead', 'RolesAllowedToWrite', 'RolesAllowedToUpdate', 'RolesAllowedToDelete'
]

SECURITY_ROLES_ATTRIBUTES = [
    'RolesAllowedToRead', 'RolesAllowedToWrite', 'RolesAllowedToUpdate', 'RolesAllowedToDelete'
]

SECURITY_IDS_ATTRIBUTES = [
    'IdsAllowedToRead', 'IdsAllowedToWrite', 'IdsAllowedToUpdate', 'IdsAllowedToDelete',
]

READ_ONLY_FIELDS = ['CreatedBy', 'CreateDate', 'LastUpdateDate', 'LastUpdateBy']

# bundled default models available in django_pds

DOCUMENT_ENTITY = 'Entity'
DOCUMENT_ROLE = 'Role'
DOCUMENT_ENTITY_DEFAULT_PERMISSION_SETTING = 'EntityDefaultPermissionSetting'
DOCUMENT_USER = 'User'
DOCUMENT_USER_READABLE_DATA = 'UserReadableData'
DOCUMENT_USER_ROLE_MAPS = 'UserRoleMap'

# JWT Token configurations

# JWT_TOKEN_SECRET_KEY add this on your custom settings
JWT_TOKEN_EXPIRATION_TIME = 6  # 6 min
JWT_TOKEN_AUDIENCE = '*'
JWT_TOKEN_ALGORITHM = 'HS256'

SELECT_NOT_ALLOWED_ENTITIES = [
    'User', 'UserReadableData', 'Tenant', 'Entity',
    'EntityDefaultPermissionSetting', 'Role', 'UserReadableData', 'UserRoleMap'
]

READ_NOT_ALLOWED_ATTRIBUTES = [
    'CreatedBy', 'LastUpdateBy', 'IdsAllowedToRead', 'IdsAllowedToWrite', 'IdsAllowedToUpdate',
    'IdsAllowedToDelete', 'RolesAllowedToRead', 'RolesAllowedToWrite', 'RolesAllowedToUpdate', 'RolesAllowedToDelete'
]

MONGO_ENGINE_USABLE_OPERATORS = [
    'ne', 'lt', 'lte', 'gt', 'in', 'nin', 'mod', 'all', 'size', 'exists', 'exact', 'iexact', 'contains', 'icontains',
    'startswith', 'istartswith', 'endswith', 'iendswith', 'match'
]

EDIT_NOT_ALLOWED_ATTRIBUTES_PDS = [
    'CreatedBy', 'CreateDate', 'LastUpdateDate', 'LastUpdateBy', 'IdsAllowedToRead', 'IdsAllowedToWrite',
    'IdsAllowedToUpdate', 'IdsAllowedToDelete', 'RolesAllowedToRead', 'RolesAllowedToWrite', 'RolesAllowedToUpdate',
    'RolesAllowedToDelete'
]



from django.conf import settings


class Settings(object):

    @property
    def SESSION_MIDDLEWARE_EXEMPT_URLS(self):
        return getattr(settings, 'TOKEN_EXEMPT_URLS', [])

    @property
    def AUTH_NOT_REQUIRED_URLS(self):
        return getattr(settings, 'AUTH_NOT_REQUIRED_URLS', [])


conf = Settings()



"""
Settings and configuration for django_pds.

this file is re-created from django.conf.__init__.py file
main reason of this settings to lazy load all the configuration
from either django_pds.core.settings file or to load
settings for django_pds defined in django project settings
"""

import importlib

from django.utils.functional import LazyObject, empty
from mongoengine import ImproperlyConfigured

from .core.utils import get_environment

ENVIRONMENT_VARIABLE = "DJANGO_PDS_SETTINGS_MODULE"

DEFAULT_CORE_SETTINGS_MODULE = 'django_pds.core.settings'

list_settings = [
    "SYSTEM_SUPPORTED_ROLES",
    "SECURITY_ATTRIBUTES",
    "SECURITY_ROLES_ATTRIBUTES",
    "SECURITY_IDS_ATTRIBUTES",
    "READ_ONLY_FIELDS",
    "SELECT_NOT_ALLOWED_ENTITIES",
    "READ_NOT_ALLOWED_ATTRIBUTES",
    "MONGO_ENGINE_USABLE_OPERATORS",
    "EDIT_NOT_ALLOWED_ATTRIBUTES_PDS"
]


class LazySettings(LazyObject):

    def _setup(self, name=None):

        default_settings_module = DEFAULT_CORE_SETTINGS_MODULE
        self._wrapped = Settings(default_settings_module)

        override_settings_module = get_environment(ENVIRONMENT_VARIABLE, False, None)
        print(override_settings_module)
        if override_settings_module:
            override_settings = SettingsOverride(self._wrapped)
            override_settings.override(override_settings_module)

    def __repr__(self):
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return '<django_pds.conf.LazySettings [Unevaluated]>'
        return '<django_pds.conf.LazySettings "%(settings_module)s">' % {
            'settings_module': self._wrapped.SETTINGS_MODULE,
        }

    def __getattr__(self, name):
        """Return the value of a setting and cache it in self.__dict__."""
        if self._wrapped is empty:
            self._setup(name)
        val = getattr(self._wrapped, name)
        self.__dict__[name] = val
        return val

    def __setattr__(self, name, value):
        """
        Set the value of setting. Clear all cached values if _wrapped changes
        (@override_settings does this) or clear single values when set.
        """
        if name == '_wrapped':
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        """Delete a setting and clear it from cache if needed."""
        super().__delattr__(name)
        self.__dict__.pop(name, None)


class Settings:
    def __init__(self, settings_module):
        self.SETTINGS_MODULE = settings_module

        mod = importlib.import_module(self.SETTINGS_MODULE)

        self._explicit_settings = set()
        for setting in dir(mod):
            if setting.isupper():
                setting_value = getattr(mod, setting)

                if (setting in list_settings and
                        not isinstance(setting_value, (list, tuple))):
                    raise ImproperlyConfigured("The %s setting must be a list or a tuple. " % setting)
                setattr(self, setting, setting_value)
                self._explicit_settings.add(setting)


class SettingsOverride:

    def __init__(self, __settings):
        self.__settings = __settings

    def override(self, override_module):

        mod = importlib.import_module(override_module)

        for setting in dir(mod):
            if setting.isupper():
                setting_value = getattr(mod, setting)

                if (setting in list_settings and
                        not isinstance(setting_value, (list, tuple))):
                    raise ImproperlyConfigured("The %s setting must be a list or a tuple. " % setting)
                setattr(self.__settings, setting, setting_value)


settings = LazySettings()



from .entity import Entity
from .entitydefaultpermissionsettings import EntityDefaultPermissionSetting
from .role import Role
from .userreadabledata import UserReadableData
from .userrolemaps import UserRoleMap

__all__ = (
    'Entity',
    'EntityDefaultPermissionSetting',
    'Role',
    'UserReadableData',
    'UserRoleMap'
)



from .base import BaseController

USER_READABLE_DATA = "UserReadableData"


class UserReadableDataController(BaseController):

    def __get_filter(self, entity_name, roles, exclude_default):

        if roles:
            if not isinstance(roles, (list, tuple)):
                return True, 'roles must be a list or tuple type'
            if 'default' not in roles:
                if not exclude_default:
                    roles.append('default')
        else:
            roles = []
            if not exclude_default:
                roles = ['default']

        return {
            'EntityName': entity_name,
            'Role__in': roles
        }

    def get_user_readable_data_fields(self, entity_name, roles=None, exclude_default=False):
        _filter = self.__get_filter(entity_name, roles, exclude_default)
        user_readable_data_s = self.get_document(USER_READABLE_DATA).objects(**_filter)
        if user_readable_data_s.count() > 0:
            fields = []
            for urd in user_readable_data_s:
                fields.append(set(urd.UserReadableFields))
            data = set().union(*fields)
            return False, data
        return True, []



__title__ = 'Django Platform Data Service'
__version__ = '0.0.1'
__author__ = 'Komol Nath Roy'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020 Komol Nath Roy'

VERSION = __version__

default_app_config = 'django_pds.apps.DjangoPdsConfig'



import json

from django_pds.conf import settings
from django_pds.core.controllers.base import BaseController

OWNER = 'owner'


class GenericInsertCommandController(BaseController):

    def __modify_ids(self, __defaults, user_id):
        items = []
        for _id in __defaults:
            if _id == OWNER:
                items.append(user_id)
            else:
                items.append(_id)
        return items

    def json_load(self, json_string):
        try:
            return False, json.loads(json_string)
        except BaseException as e:
            return True, str(e)

    def already_exists(self, document_name, document_id):
        try:
            data = self.get_document(document_name).objects(ItemId=document_id)
            return data.count() > 0
        except BaseException as e:
            return False

    def insert_one(self, document_name, data, user_id=None, default_permission=None):
        try:

            base_instance = self.is_base_instance(document_name)
            simple_base_instance = self.is_simple_base_doc_instance(document_name)

            if not base_instance and not simple_base_instance:
                return True, 'Document type must be `BaseDocument` ' \
                             'or `SimpleBaseDocument` ' \
                             'from django_pds.core.base Module'

            Model = self.get_document(document_name)
            mod = Model(**data)

            if base_instance:

                if user_id:
                    mod.CreatedBy = user_id
                    mod.LastUpdateBy = user_id
                    for item in settings.SECURITY_IDS_ATTRIBUTES:
                        ids = default_permission.get(item, [])
                        ids = self.__modify_ids(ids, user_id)
                        setattr(mod, item, ids)

                if default_permission:
                    for item in settings.SECURITY_ROLES_ATTRIBUTES:
                        roles = default_permission.get(item, [])
                        setattr(mod, item, roles)
                    setattr(mod, 'RolesAllowedToWrite', [])
                    setattr(mod, 'IdsAllowedToWrite', [])

            mod.save()
            return False, mod.ItemId

        except BaseException as e:
            return True, e

    def insert_many(self, document_name, data_array, user_id=None, default_permission=None):
        results = []
        for data in data_array:
            err, item_id = self.insert_one(document_name, data, user_id, default_permission)
            if err:
                results.append(None)
            else:
                results.append(item_id)
        return results



from django_pds.core.controllers.base import RequiredController


def required(*fields):
    def required_wrapper(func):
        def checking_required(*args, **kwargs):
            required_ctrl = RequiredController()
            request = args[1]
            error, response = required_ctrl.required(request, *fields)
            if error:
                return response
            return func(*args, **kwargs)

        return checking_required

    return required_wrapper



from uuid import uuid4

from .base import BaseController

USER_ROLE_MAP = 'UserRoleMap'


class UserRoleMapsController(BaseController):

    def get_roles(self, user_id):
        roles = self.get_document(USER_ROLE_MAP).objects(IdsAllowedToRead=user_id)
        return roles

    def get_user_non_dynamic_roles(self, available_not_dynamic_roles, user_id):
        user_roles = self.get_user_roles(user_id)
        user_roles = set(user_roles)
        available_not_dynamic_roles = set(available_not_dynamic_roles)
        return user_roles.intersection(available_not_dynamic_roles)

    def get_user_roles(self, user_id):
        return self.get_document(USER_ROLE_MAP).objects(IdsAllowedToRead=user_id).distinct(field='RoleName')

    def create_user_role_map(self, role_name, user_id):

        try:

            UserRoleMap = self.get_document(USER_ROLE_MAP)

            role = UserRoleMap()
            role.ItemId = str(uuid4())
            role.UserId = user_id
            role.RoleName = role_name

            role.CreatedBy = "system"
            role.LastUpdateBy = "system"

            role.IdsAllowedToRead = [user_id]
            role.IdsAllowedToUpdate = [user_id]
            role.IdsAllowedToWrite = [user_id]

            role.save()

            return False, role.ItemId

        except BaseException as e:
            return True, e



from .base import BaseController

ROLE = 'Role'


class RoleController(BaseController):

    def get_non_dynamic_roles(self):
        return self.get_document(ROLE).objects(IsDynamic=False).distinct(field='RoleName')

    def get_roles(self):
        pass

    def is_dynamic_role(self):
        pass

    def create_role(self, role_name, is_dynamic=True):
        pass



import json
from json.decoder import JSONDecodeError


class xstr(str):
    """
    xstr is a subclass of string (str)
    converts string to any other types
    """

    def isbool(self):
        return self == 'True' or self == 'true' or self == 'False' or self == 'false'

    def convert_to_bool(self):
        if self == 'True' or self == 'true':
            return True
        return False

    def isint(self):
        try:
            item = int(self)
            return True
        except ValueError:
            return False

    def isfloat(self):
        try:
            item = float(self)
            return True
        except ValueError:
            return False

    def is_dict(self):
        try:
            item = json.loads(self)
            return True
        except JSONDecodeError:
            return False

    def get(self):

        if self.isbool():
            return self.convert_to_bool()

        if self.isint():
            return int(self)

        if self.isfloat():
            return float(self)

        if self.is_dict():
            return json.loads(self)

        if self.startswith('\'') and self.endswith('\''):
            return self[1:len(self) - 1]

        return self.strip()
